import minidb

from nose.tools import *

import datetime


class FieldTest(minidb.Model):
    CONSTANT = 123

    # Persisted
    column1 = str
    column2 = int
    column3 = float
    column4 = bool

    # Not persisted per-instance attribute
    _private1 = object
    _private2 = str
    _private3 = int
    _private4 = object

    # Class attributes
    __class_attribute1__ = 'Hello'
    __class_attribute2__ = ['World']

    def __init__(self, constructor_arg):
        self._private1 = constructor_arg
        self._private2 = 'private'
        self._private3 = self.CONSTANT
        self._private4 = None

    @classmethod
    def a_classmethod(cls):
        return 'classmethod'

    @staticmethod
    def a_staticmethod():
        return 'staticmethod'

    def a_membermethod(self):
        return self._private1

    @property
    def a_read_only_property(self):
        return self._private2.upper()

    @property
    def a_read_write_property(self):
        return self._private3

    @a_read_write_property.setter
    def read_write_property(self, new_value):
        self._private3 = new_value


class FieldConversion(minidb.Model):
    integer = int
    floating = float
    boolean = bool
    string = str
    jsoninteger = minidb.JSON
    jsonfloating = minidb.JSON
    jsonboolean = minidb.JSON
    jsonstring = minidb.JSON
    jsonlist = minidb.JSON
    jsondict = minidb.JSON
    jsonnone = minidb.JSON

    @classmethod
    def create(cls):
        return cls(integer=1, floating=1.1, boolean=True, string='test',
                   jsonlist=[1, 2], jsondict={'a': 1}, jsonnone=None,
                   jsoninteger=1, jsonfloating=1.1, jsonboolean=True,
                   jsonstring='test')


def test_instantiate_fieldtest_from_code():
    field_test = FieldTest(999)
    assert field_test.id is None
    assert field_test.column1 is None
    assert field_test.column2 is None
    assert field_test.column3 is None
    assert field_test.column4 is None
    assert field_test._private1 == 999
    assert field_test._private2 is not None
    assert field_test._private3 is not None
    assert field_test._private4 is None


def test_saving_object_stores_id():
    with minidb.Store(debug=True) as db:
        db.register(FieldTest)
        field_test = FieldTest(998)
        assert field_test.id is None
        field_test.save(db)
        assert field_test.id is not None


def test_loading_object_returns_cached_object():
    with minidb.Store(debug=True) as db:
        db.register(FieldTest)
        field_test = FieldTest(9999)
        field_test._private1 = 4711
        assert field_test.id is None
        field_test.save(db)
        assert field_test.id is not None
        field_test_loaded = FieldTest.get(db, id=field_test.id)(9999)
        assert field_test_loaded._private1 == 4711
        assert field_test_loaded is field_test


def test_loading_object_returns_new_object_after_reference_drop():
    with minidb.Store(debug=True) as db:
        db.register(FieldTest)
        field_test = FieldTest(9999)
        field_test._private1 = 4711
        assert field_test.id is None
        field_test.save(db)
        assert field_test.id is not None
        field_test_id = field_test.id
        del field_test

        field_test_loaded = FieldTest.get(db, id=field_test_id)(9999)
        assert field_test_loaded._private1 == 9999


def test_loading_objects():
    with minidb.Store(debug=True) as db:
        db.register(FieldTest)
        for i in range(100):
            FieldTest(i).save(db)

        assert next(FieldTest.c.id.count('count').query(db)).count == 100

        for field_test in FieldTest.load(db)(997):
            assert field_test.id is not None
            assert field_test._private1 == 997


@raises(minidb.UnknownClass)
def test_saving_without_registration_fails():
    with minidb.Store(debug=True) as db:
        FieldTest(9).save(db)


@raises(TypeError)
def test_registering_non_subclass_of_model_fails():
    # This cannot be registered, as it's not a subclass of minidb.Model
    class Something(object):
        column = str

    with minidb.Store(debug=True) as db:
        db.register(Something)
        db.register(Something)


@raises(KeyError)
def test_invalid_keyword_arguments_fails():
    with minidb.Store(debug=True) as db:
        db.register(FieldTest)
        FieldTest(9, this_is_not_an_attribute=123).save(db)


@raises(AttributeError)
def test_invalid_column_raises_attribute_error():
    class HasOnlyColumnX(minidb.Model):
        x = int

    with minidb.Store(debug=True) as db:
        db.register(HasOnlyColumnX)
        HasOnlyColumnX.c.y


def test_json_serialization():
    class WithJsonField(minidb.Model):
        foo = str
        bar = minidb.JSON

    with minidb.Store(debug=True) as db:
        db.register(WithJsonField)
        d = {'a': 1, 'b': [1, 2, 3], 'c': [True, 4.0, {'d': 'e'}]}
        WithJsonField(bar=d).save(db)
        assert WithJsonField.get(db, id=1).bar == d


def test_json_field_query():
    class WithJsonField(minidb.Model):
        bar = minidb.JSON

    with minidb.Store(debug=True) as db:
        db.register(WithJsonField)
        d = {'a': [1, True, 3.9]}
        WithJsonField(bar=d).save(db)
        eq_(next(WithJsonField.c.bar.query(db)).bar, d)


def test_json_field_renamed_query():
    class WithJsonField(minidb.Model):
        bar = minidb.JSON

    with minidb.Store(debug=True) as db:
        db.register(WithJsonField)
        d = {'a': [1, True, 3.9]}
        WithJsonField(bar=d).save(db)
        eq_(next(WithJsonField.c.bar('renamed').query(db)).renamed, d)


def test_field_conversion_get_object():
    with minidb.Store(debug=True) as db:
        db.register(FieldConversion)
        FieldConversion.create().save(db)
        result = FieldConversion.get(db, id=1)
        assert isinstance(result.integer, int)
        assert isinstance(result.floating, float)
        assert isinstance(result.boolean, bool)
        assert isinstance(result.string, str)
        assert isinstance(result.jsoninteger, int)
        assert isinstance(result.jsonfloating, float)
        assert isinstance(result.jsonboolean, bool)
        assert isinstance(result.jsonstring, str)
        assert isinstance(result.jsonlist, list)
        assert isinstance(result.jsondict, dict)
        assert result.jsonnone is None


def test_field_conversion_query_select_star():
    with minidb.Store(debug=True) as db:
        db.register(FieldConversion)
        FieldConversion.create().save(db)
        result = next(FieldConversion.query(db, minidb.literal('*')))
        assert isinstance(result.integer, int)
        assert isinstance(result.floating, float)
        assert isinstance(result.boolean, bool)
        assert isinstance(result.string, str)
        assert isinstance(result.jsoninteger, int)
        assert isinstance(result.jsonfloating, float)
        assert isinstance(result.jsonboolean, bool)
        assert isinstance(result.jsonstring, str)
        assert isinstance(result.jsonlist, list)
        assert isinstance(result.jsondict, dict)
        assert result.jsonnone is None


def test_storing_and_retrieving_booleans():
    class BooleanModel(minidb.Model):
        value = bool

    with minidb.Store(debug=True) as db:
        db.register(BooleanModel)
        true_id = BooleanModel(value=True).save(db).id
        false_id = BooleanModel(value=False).save(db).id
        assert BooleanModel.get(db, id=true_id).value is True
        assert BooleanModel.get(db, BooleanModel.c.id == true_id).value is True
        assert BooleanModel.get(db, lambda c: c.id == true_id).value is True
        assert BooleanModel.get(db, id=false_id).value is False
        assert next(BooleanModel.c.value.query(db, where=lambda c: c.id == true_id)).value is True
        assert next(BooleanModel.c.value.query(db, where=lambda c: c.id == false_id)).value is False


def test_storing_and_retrieving_floats():
    class FloatModel(minidb.Model):
        value = float

    with minidb.Store(debug=True) as db:
        db.register(FloatModel)
        float_id = FloatModel(value=3.1415).save(db).id
        get_value = FloatModel.get(db, id=float_id).value
        assert type(get_value) == float
        assert get_value == 3.1415
        query_value = next(FloatModel.c.value.query(db, where=lambda c: c.id == float_id)).value
        assert type(query_value) == float
        assert query_value == 3.1415


def test_storing_and_retrieving_bytes():
    # http://probablyprogramming.com/2009/03/15/the-tiniest-gif-ever
    BLOB = (b'GIF89a\x01\x00\x01\x00\x80\x01\x00\xff\xff\xff\x00\x00\x00' +
            b'!\xf9\x04\x01\n\x00\x01\x00,\x00\x00\x00\x00\x01\x00\x01' +
            b'\x00\x00\x02\x02L\x01\x00;')

    class BytesModel(minidb.Model):
        value = bytes

    with minidb.Store(debug=True) as db:
        db.register(BytesModel)
        bytes_id = BytesModel(value=BLOB).save(db).id
        get_value = BytesModel.get(db, id=bytes_id).value
        assert type(get_value) == bytes
        assert get_value == BLOB
        query_value = next(BytesModel.c.value.query(db, where=lambda c: c.id == bytes_id)).value
        assert type(query_value) == bytes
        assert query_value == BLOB


@raises(ValueError)
def test_get_with_multiple_value_raises_exception():
    class Mod(minidb.Model):
        mod = str

    with minidb.Store(debug=True) as db:
        db.register(Mod)
        Mod(mod='foo').save(db)
        Mod(mod='foo').save(db)
        Mod.get(db, mod='foo')


def test_get_with_no_value_returns_none():
    class Mod(minidb.Model):
        mod = str

    with minidb.Store(debug=True) as db:
        db.register(Mod)
        assert Mod.get(db, mod='foo') is None


def test_delete_where():
    class DeleteWhere(minidb.Model):
        v = int

    with minidb.Store(debug=True) as db:
        db.register(DeleteWhere)

        for i in range(10):
            DeleteWhere(v=i).save(db)

        assert DeleteWhere.delete_where(db, lambda c: c.v < 2) == len({0, 1})
        assert DeleteWhere.delete_where(db, DeleteWhere.c.v > 5) == len({6, 7, 8, 9})
        assert {2, 3, 4, 5} == {v for (v,) in DeleteWhere.c.v.query(db)}


@raises(AttributeError)
def test_invalid_rowproxy_access_by_attribute():
    class Foo(minidb.Model):
        bar = str

    with minidb.Store(debug=True) as db:
        db.register(Foo)
        Foo(bar='baz').save(db)
        next(Foo.query(db, Foo.c.bar)).baz


@raises(KeyError)
def test_invalid_rowproxy_access_by_key():
    class Foo(minidb.Model):
        bar = str

    with minidb.Store(debug=True) as db:
        db.register(Foo)
        Foo(bar='baz').save(db)
        next(Foo.query(db, Foo.c.bar))['baz']


@raises(TypeError)
def test_use_schema_without_registration_raises_typeerror():
    with minidb.Store(debug=True) as db:
        class Foo(minidb.Model):
            bar = str
        Foo.query(db)


@raises(TypeError)
def test_use_schema_with_nonidentity_class_raises_typeerror():
    with minidb.Store(debug=True) as db:
        class Foo(minidb.Model):
            bar = str
        db.register(Foo)

        class Foo(minidb.Model):
            bar = str

        Foo.query(db)


@raises(TypeError)
def test_upgrade_schema_without_upgrade_raises_typeerror():
    with minidb.Store(debug=True) as db:
        class Foo(minidb.Model):
            bar = str

        db.register(Foo)

        class Foo(minidb.Model):
            bar = str
            baz = int

        db.register(Foo)


@raises(TypeError)
def test_reregistering_class_raises_typeerror():
    class Foo(minidb.Model):
        bar = int

    with minidb.Store(debug=True) as db:
        db.register(Foo)
        db.register(Foo)


def test_upgrade_schema_with_upgrade_succeeds():
    with minidb.Store(debug=True) as db:
        class Foo(minidb.Model):
            bar = str

        db.register(Foo)

        class Foo(minidb.Model):
            bar = str
            baz = int

        db.register(Foo, upgrade=True)


@raises(TypeError)
def test_upgrade_schema_with_different_type_raises_typeerror():
    with minidb.Store(debug=True) as db:
        class Foo(minidb.Model):
            bar = str

        db.register(Foo)

        class Foo(minidb.Model):
            bar = int

        db.register(Foo, upgrade=True)


def test_update_object():
    class Foo(minidb.Model):
        bar = str

    with minidb.Store(debug=True) as db:
        db.register(Foo)
        a = Foo(bar='a').save(db)
        b = Foo(bar='b').save(db)

        a.bar = 'c'
        a.save()

        b.bar = 'd'
        b.save()

        assert {'c', 'd'} == {bar for (bar,) in Foo.c.bar.query(db)}


def test_delete_object():
    class Foo(minidb.Model):
        bar = int

    with minidb.Store(debug=True) as db:
        db.register(Foo)
        for i in range(3):
            Foo(bar=i).save(db)

        Foo.get(db, bar=2).delete()

        assert {0, 1} == {bar for (bar,) in Foo.c.bar.query(db)}


def test_distinct():
    class Foo(minidb.Model):
        bar = str
        baz = int

    with minidb.Store(debug=True) as db:
        db.register(Foo)

        for i in range(2):
            Foo(bar='hi', baz=i).save(db)

        Foo(bar='ho', baz=7).save(db)

        expected = {('hi',), ('ho',)}

        # minidb.func.distinct(COLUMN)(NAME)
        result = {tuple(x) for x in Foo.query(db, lambda c: minidb.func.distinct(c.bar)('foo'))}
        eq_(result, expected)

        # COLUMN.distinct(NAME)
        result = {tuple(x) for x in Foo.query(db, Foo.c.bar.distinct('foo'))}
        eq_(result, expected)



def test_group_by_with_sum():
    class Foo(minidb.Model):
        bar = str
        baz = int

    with minidb.Store(debug=True) as db:
        db.register(Foo)

        for i in range(5):
            Foo(bar='hi', baz=i).save(db)

        for i in range(6):
            Foo(bar='ho', baz=i).save(db)

        expected = {('hi', sum(range(5))), ('ho', sum(range(6)))}

        # minidb.func.sum(COLUMN)(NAME)
        result = {tuple(x) for x in Foo.query(db, lambda c: c.bar //
                  minidb.func.sum(c.baz)('sum'), group_by=lambda c: c.bar)}
        eq_(result, expected)

        # COLUMN.sum(NAME)
        result = {tuple(x) for x in Foo.query(db, lambda c: c.bar //
                  c.baz.sum('sum'), group_by=lambda c: c.bar)}
        eq_(result, expected)


@raises(ValueError)
def test_save_without_db_raises_valueerror():
    class Foo(minidb.Model):
        bar = int

    Foo(bar=99).save()


@raises(ValueError)
def test_delete_without_db_raises_valueerror():
    class Foo(minidb.Model):
        bar = int

    Foo(bar=99).delete()


@raises(KeyError)
def test_double_delete_without_id_raises_valueerror():
    class Foo(minidb.Model):
        bar = str

    with minidb.Store(debug=True) as db:
        db.register(Foo)
        a = Foo(bar='hello')
        a.save(db)
        assert a.id is not None
        a.delete()
        assert a.id is None
        a.delete()


def test_default_values_are_set_if_none():
    class Foo(minidb.Model):
        name = str

        class __minidb_defaults__:
            name = 'Bob'

    with minidb.Store(debug=True) as db:
        f = Foo()
        eq_(f.name, 'Bob')

        f = Foo(name='John')
        eq_(f.name, 'John')


def test_default_values_with_callable():
    class Foo(minidb.Model):
        name = str
        email = str

        # Defaults are applied in order of slots of the Model
        # subclass, so if e.g. email depends on name to be
        # set, make sure email appears *after* name in the model
        class __minidb_defaults__:
            name = lambda o: 'Bob'
            email = lambda o: o.name + '@example.com'

    with minidb.Store(debug=True) as db:
        f = Foo()
        eq_(f.name, 'Bob')
        eq_(f.email, 'Bob@example.com')

        f = Foo(name='John')
        eq_(f.name, 'John')
        eq_(f.email, 'John@example.com')

        f = Foo(name='Joe', email='joe@example.net')
        eq_(f.name, 'Joe')
        eq_(f.email, 'joe@example.net')


def test_storing_and_retrieving_datetime():
    DT_NOW = datetime.datetime.now()
    D_TODAY = datetime.date.today()
    T_NOW = datetime.datetime.now().time()

    class DateTimeModel(minidb.Model):
        dt = datetime.datetime
        da = datetime.date
        tm = datetime.time

    with minidb.Store(debug=True) as db:
        db.register(DateTimeModel)
        datetime_id = DateTimeModel(dt=DT_NOW, da=D_TODAY, tm=T_NOW).save(db).id
        get_value = DateTimeModel.get(db, id=datetime_id)
        assert type(get_value.dt) == datetime.datetime
        assert get_value.dt == DT_NOW
        assert type(get_value.da) == datetime.date
        assert get_value.da == D_TODAY
        assert type(get_value.tm) == datetime.time
        assert get_value.tm == T_NOW
        query_value = next(DateTimeModel.query(db, lambda c: c.dt // c.da // c.tm,
                                               where=lambda c: c.id == datetime_id))
        assert type(query_value.dt) == datetime.datetime
        assert query_value.dt == DT_NOW
        assert type(query_value.da) == datetime.date
        assert query_value.da == D_TODAY
        assert type(query_value.tm) == datetime.time
        assert query_value.tm == T_NOW


def test_query_with_datetime():
    DT_NOW = datetime.datetime.now()

    class DateTimeModel(minidb.Model):
        dt = datetime.datetime

    with minidb.Store(debug=True) as db:
        db.register(DateTimeModel)
        datetime_id = DateTimeModel(dt=DT_NOW).save(db).id
        assert DateTimeModel.get(db, lambda c: c.dt == DT_NOW).id == datetime_id


def test_custom_converter():
    class Point(object):
        def __init__(self, x, y):
            self.x = x
            self.y = y

    @minidb.converter_for(Point)
    def convert_point(v, serialize):
        if serialize:
            return ','.join(str(x) for x in (v.x, v.y))
        else:
            return Point(*(float(x) for x in v.split(',')))

    class Player(minidb.Model):
        name = str
        position = Point

    with minidb.Store(debug=True) as db:
        db.register(Player)
        p = Point(1.12, 5.99)
        player_id = Player(name='Foo', position=p).save(db).id
        get_value = Player.get(db, id=player_id)
        assert type(get_value.position) == Point
        assert (get_value.position.x, get_value.position.y) == (p.x, p.y)
        query_value = next(Player.query(db, lambda c: c.position,
                                        where=lambda c: c.id == player_id))
        assert type(query_value.position) == Point
        assert (query_value.position.x, query_value.position.y) == (p.x, p.y)



import minidb


class Person(minidb.Model):
    # Constants have to be all-uppercase
    THIS_IS_A_CONSTANT = 123

    # Database columns have to be lowercase
    username = str
    mail = str
    foo = int

    # Not persisted (runtime-only) class attibutes start with underscore
    _not_persisted = float
    _foo = object

    # Custom, non-constant class attributes with dunder
    __custom_class_attribute__ = []

    # This is the custom constructor that will be called by minidb.
    # The attributes from the db will already be set when this is called.
    def __init__(self, foo):
        print('here we go now:', foo, self)
        self._not_persisted = 42.23
        self._foo = foo

    @classmethod
    def cm_foo(cls):
        print('called classmethod', cls)

    @staticmethod
    def sm_bar():
        print('called static method')

    def send_email(self):
        print('Would send e-mail to {self.username} at {self.mail}'.format(self=self))
        print('and we have _foo as', self._foo)

    @property
    def a_property(self):
        return self.username.upper()

    @property
    def read_write_property(self):
        return 'old value' + str(self.THIS_IS_A_CONSTANT)

    @read_write_property.setter
    def read_write_property(self, new):
        print('new value:', new)


class AdvancedPerson(Person):
    advanced_x = float
    advanced_y = float


class WithoutConstructor(minidb.Model):
    name = str
    age = int
    height = float


class WithPayload(minidb.Model):
    payload = minidb.JSON


Person.__custom_class_attribute__.append(333)
print(Person.__custom_class_attribute__)
Person.cm_foo()
Person.sm_bar()


class FooObject(object):
    pass


with minidb.Store(debug=True) as db:
    db.register(Person)
    db.register(WithoutConstructor)
    db.register(AdvancedPerson)
    db.register(WithPayload)

    AdvancedPerson(username='advanced', mail='a@example.net').save(db)

    for aperson in AdvancedPerson.load(db):
        print(aperson)

    for i in range(5):
        w = WithoutConstructor(name='x', age=10 + 3 * i)
        w.height = w.age * 3.33
        w.save(db)
        print(w)
        w2 = WithoutConstructor()
        w2.name = 'xx'
        w2.age = 100 + w.age
        w2.height = w2.age * 23.33
        w2.save(db)
        print(w2)

    for i in range(3):
        p = Person(FooObject(), username='foo' * i)
        print(p)
        p.save(db)
        print(p)
        p.username *= 3
        p.save()
        pp = Person(FooObject())
        pp.username = 'bar' * i
        print(pp)
        pp.save(db)
        print(pp)

    print('loader is:', Person.load(db))

    print('query')
    # for person in db.load(Person, FooObject()):
    for person in Person.load(db)(FooObject()):
        print(person)
        if person.username == '':
            print('delete')
            person.delete()
            print('id after delete:', person.id)
            continue
        person.mail = person.username + '@example.com'
        person.save()
        print(person)

    print('query without')
    for w in WithoutConstructor.load(db):
        print(w)

    print('get without')
    w = WithoutConstructor.get(db, age=13)
    print('got:', w)

    print('requery')
    print({p.id: p for p in Person.load(db)(FooObject())})
    person = Person.get(db, id=3)(FooObject())
    # person = db.get(Person, FooObject(), id=2)
    print(person)
    person.send_email()
    print('a_property:', person.a_property)
    print('rw property:', person.read_write_property)
    person.read_write_property = 'hello'
    print('get not persisted:', person._not_persisted)
    person._not_persisted = 47.11
    print(person._not_persisted)
    person.save()

    print('RowProxy')
    for row in Person.query(db, Person.c.username // Person.c.foo):
        print('Repr:', row)
        print('Attribute access:', row.username, row.foo)
        print('Key access:', row['username'], row['foo'])
        print('Index access:', row[0], row[1])
        print('As dict:', dict(row))

    print('select with query builder')
    print('columns:', Person.c)
    query = (Person.c.id < 1000) & Person.c.username.like('%foo%') & (Person.c.username != None)
    # Person.load(db, Person.id < 1000 & Person.username.like('%foo%'))
    print('query:', query)
    print({p.id: p for p in Person.load(db, query)(FooObject())})
    print('deleting all persons with a short username')
    print(Person.delete_where(db, Person.c.username.length <= 3))

    print('what is left')
    for p in Person.load(db)(FooObject()):
        uu = next(Person.query(db, minidb.columns(Person.c.username.upper('up'),
                                                  Person.c.username.lower('down'),
                                                  Person.c.foo('foox'),
                                                  Person.c.foo),
                               where=(Person.c.id == p.id),
                               order_by=minidb.columns(Person.c.id.desc,
                                                       Person.c.username.length.asc),
                               limit=1))
        print(p.id, p.username, p.mail, uu)

    print('=' * 30)
    print('queries')
    print('=' * 30)

    highest_id = next(Person.query(db, Person.c.id.max('max'))).max
    print('highest id:', highest_id)

    average_age = next(WithoutConstructor.query(db, WithoutConstructor.c.age.avg('average'))).average
    print('average age:', average_age)

    all_ages = list(WithoutConstructor.c.age.query(db, order_by=WithoutConstructor.c.age.desc))
    print('all ages:', all_ages)

    average_age = next(WithoutConstructor.c.age.avg('average').query(db, limit=1)).average
    print('average age (direct query):', average_age)

    print('multi-column query:')
    for row in WithoutConstructor.query(db, minidb.columns(WithoutConstructor.c.age,
                                                           WithoutConstructor.c.height),
                                        order_by=WithoutConstructor.c.age.desc,
                                        limit=50):
        print('got:', dict(row))

    print('multi-column query (direct)')
    print([dict(x) for x in minidb.columns(WithoutConstructor.c.age,
                                           WithoutConstructor.c.height).query(
                                               db, order_by=WithoutConstructor.c.height.desc)])

    print('order by multiple with then')
    print(list(WithoutConstructor.c.age.query(db, order_by=(WithoutConstructor.c.height.asc //
                                                            WithoutConstructor.c.age.desc))))

    print('order by shortcut with late-binding column lambda as dictionary')
    print(list(WithoutConstructor.c.age.query(db, order_by=lambda c: c.height.asc // c.age.desc)))

    print('multiple columns with // and as tuple')
    for age, height in (WithoutConstructor.c.age // WithoutConstructor.c.height).query(db):
        print(age, height)

    print('simple query for age')
    for (age,) in WithoutConstructor.c.age.query(db):
        print(age)

    print('late-binding column lambda')
    for name, age, height, random in WithoutConstructor.query(db, lambda c: (c.name // c.age //
                                                                             c.height // minidb.func.random()),
                                                              order_by=lambda c: (c.height.desc //
                                                                                  minidb.func.random().asc)):
        print('got:', name, age, height, random)

    print(minidb.func.max(1, Person.c.username, 3, minidb.func.random()).tosql())
    print(minidb.func.max(Person.c.username.lower, person.c.foo.lower, 6)('maximal').tosql())

    print('...')
    print(Person.load(db, Person.c.username.like('%'))(FooObject()))

    print('Select Star')
    print(list(Person.query(db, minidb.literal('*'))))

    print('Group By')
    print(list(Person.query(db, Person.c.username // Person.c.id.count, group_by=Person.c.username)))

    print('Pretty-Printing')
    minidb.pprint(Person.query(db, minidb.literal('*')))

    print('Pretty-formatting in color')
    print(repr(minidb.pformat(Person.query(db, Person.c.id), color=True)))

    print('Pretty-Printing, in color')
    minidb.pprint(WithoutConstructor.query(db, minidb.literal('*')), color=True)

    print('Pretty-Querying with default star-select')
    Person.pquery(db)

    print('With payload (JSON)')
    WithPayload(payload={'a': [1] * 3}).save(db)
    for payload in WithPayload.load(db):
        print('foo', payload)

    print(next(WithPayload.c.payload.query(db)))


def cached_person_main(with_delete=None):
    if with_delete is None:
        for i in range(2):
            cached_person_main(i)
        print('=' * 77)
        return
    print('=' * 20, 'Cached Person Main, with_delete =', with_delete, '=' * 20)

    debug_object_cache, minidb.DEBUG_OBJECT_CACHE = minidb.DEBUG_OBJECT_CACHE, True

    class CachedPerson(minidb.Model):
        name = str
        age = int
        _inst = object

    with minidb.Store(debug=True) as db:
        db.register(CachedPerson)
        p = CachedPerson(name='foo', age=12)
        p._inst = 123
        p.save(db)
        p_id = p.id
        if with_delete:
            del p
        p = CachedPerson.get(db, id=p_id)
        print('p._inst =', repr(p._inst))
    minidb.DEBUG_OBJECT_CACHE = debug_object_cache

cached_person_main()



#!/usr/bin/env python3
# Setup script for
'minidb'
# by Thomas Perl <thp.io>

import os
import re

dirname = os.path.dirname(os.path.abspath(__file__))
src = open(os.path.join(dirname, '{}.py'.format(__doc__))).read()
docstrings = re.findall('"""(.*)"""', src)
m = dict(re.findall("__([a-z_]+)__\s*=\s*'([^']+)'", src))
m['name'] = __doc__
m['author'], m['author_email'] = re.match(r'(.*) <(.*)>', m['author']).groups()
m['description'] = docstrings[0]
m['py_modules'] = (m['name'],)
m['download_url'] = '{m[url]}{m[name]}-{m[version]}.tar.gz'.format(m=m)

from distutils.core import setup
setup(**m)
